/*
 *	Real Time Clock interface for Linux on Atmel AT91RM9200
 *
 *	Copyright (C) 2002 Rick Bronson
 *
 *	Ported to Linux 2.6 by Steven Scholz
 *	Based on s3c2410-rtc.c Simtec Electronics
 *
 *      Based on sa1100-rtc.c by Nils Faerber
 *	Based on rtc.c by Paul Gortmaker
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 *     20090709 By Ace Yang Version 1.0.1
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/time.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>

#include <asm/uaccess.h>
#include <asm/rtc.h>

#include <asm/mach/time.h>

#define DRIVER_NAME "matrix500_rtc"
#define VERSION "1.0.2"

#define rtc_epoch		1900UL	/* just like arch/arm/common/rtctime.c */

#ifdef CONFIG_MATRIX
/* define for I2C rs372a driver call*/
	#define RS372A_GET_DATE		0
	#define RS372A_SET_DATE		1
	extern int rs372a_do_command(int bus, int cmd, void *arg);
#else
	#define DS1338_GET_DATE		0
	#define DS1338_SET_DATE		1
	#define DS1338_GET_SRAM		2
	#define DS1338_SET_SRAM		3
	extern int ds1338_do_command(int bus, int cmd, void *arg);
#endif


static DECLARE_WAIT_QUEUE_HEAD(matrix500_rtc_update);

extern spinlock_t rtc_lock;
static unsigned long rtc_status = 0;


/*
 * A few helper functions ...
 */

static const unsigned char days_in_mo[] =
    { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

#define is_leap(year) \
	((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))

static const unsigned short int __mon_yday[2][13] =
{
	/* Normal years. */
	{ 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
	/* Leap years. */
	{ 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
};

/*
 * Returns day since start of the year [0-365]
 *  (from drivers/char/efirtc.c)
 */
static inline int compute_yday(int year, int month, int day)
{
	return __mon_yday[is_leap(year)][month] + day-1;
}


/*
 * Read current time and date in RTC
 */
static int matrix500_rtc_readtime(struct rtc_time *tm)
{
	int ret , i = 0;
	struct rtc_time tm_buf;
	spin_lock(&rtc_lock);

#ifdef CONFIG_MATRIX
	do {
		ret = rs372a_do_command(0,RS372A_GET_DATE, &tm_buf);
		if(ret != 0 ) {spin_unlock(&rtc_lock); return ret;}

		ret = rs372a_do_command(0,RS372A_GET_DATE, tm);
		if(ret != 0 ) {spin_unlock(&rtc_lock); return ret;}
		
		if(i > 10) {spin_unlock(&rtc_lock); return -10;}
		i++;
	}while(tm->tm_year != tm_buf.tm_year || 
		tm->tm_mon != tm_buf.tm_mon ||
		tm->tm_mday != tm_buf.tm_mday ||
		tm->tm_hour != tm_buf.tm_hour ||
		tm->tm_min != tm_buf.tm_min );
	
	
#else
	ret = ds1338_do_command(0,DS1338_GET_DATE, tm);
#endif
	tm->tm_year -= rtc_epoch;
	tm->tm_mon--;
	spin_unlock(&rtc_lock);
	if(ret != 0 ) return ret;

	pr_debug("%s(): %4d-%02d-%02d %02d:%02d:%02d\n", __FUNCTION__,
		 tm->tm_year, tm->tm_mon, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);

	return 0;
}

/*
 * Set current time and date in RTC
 */
static int matrix500_rtc_settime(struct rtc_time *tm)
{
	int ret;

	spin_lock(&rtc_lock);
	tm->tm_year += rtc_epoch;
	tm->tm_mon++;

#ifdef CONFIG_MATRIX
	ret = rs372a_do_command(0, RS372A_SET_DATE, tm);
#else
	ret = ds1338_do_command(0, DS1338_SET_DATE, tm);
#endif
	spin_unlock(&rtc_lock);
	if(ret != 0) return ret;

	pr_debug("%s(): %4d-%02d-%02d %02d:%02d:%02d\n", __FUNCTION__,
		1900 + tm->tm_year, tm->tm_mon, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);

	

	return 0;
}

/*
 * Read alarm time and date in RTC
 */


/*
 * Set alarm time and date in RTC
 */


/*
 * Handle commands from user-space
 */
static int matrix500_rtc_ioctl(unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	struct rtc_time tm;

	pr_debug("%s(): cmd=%08x, arg=%08lx.\n", __FUNCTION__, cmd, arg);

//	printk(KERN_INFO "matrix500_rtc_ioctl:%08x\n",cmd);
	switch (cmd) {
	case RTC_AIE_OFF:	/* alarm off */
		
		break;
	case RTC_AIE_ON:	/* alarm on */
		
		break;
	case RTC_UIE_OFF:	/* update off */
		
		break;
	case RTC_UIE_ON:	/* update on */
		
		break;
	case RTC_PIE_OFF:	/* periodic off */
		
		break;
	case RTC_PIE_ON:	/* periodic on */
		
		break;
	case RTC_IRQP_READ:	/* read periodic alarm frequency */
		
		break;
	case RTC_IRQP_SET:	/* set periodic alarm frequency */
		
		break;
	case RTC_RD_TIME:
		

		memset(&tm, 0, sizeof (struct rtc_time));
		matrix500_rtc_readtime(&tm);
		if (copy_to_user((struct rtc_time*)arg, &tm, sizeof(struct rtc_time)))
				return -EFAULT;
		return 0;
	break;
	case RTC_SET_TIME:
		
		if (copy_from_user(&tm, (struct rtc_time*)arg, sizeof(struct rtc_time)))
			return -EFAULT;

		tm.tm_year =+ 1900;
		tm.tm_mon =+ 1;   /* tm_mon starts at zero */
						
		matrix500_rtc_settime(&tm);	
		return 0;
	break;
	default:
		ret = -EINVAL;
		break;
	}
	

	return ret;
}

/*
 * Provide additional RTC information in /proc/driver/rtc
 */
static int matrix500_rtc_read_proc(char *buf)
{
	char *p = buf;

	p += sprintf(p, "Aritla Matrix500 RTC\n");
	

	return p - buf;
}



static struct rtc_ops rtc_ops = {
	.owner		= THIS_MODULE,
	.ioctl		= matrix500_rtc_ioctl,
	.read_time	= matrix500_rtc_readtime,
	.set_time	= matrix500_rtc_settime,
	/*.read_alarm	= matrix500_rtc_readalarm,
	.set_alarm	= matrix500_rtc_setalarm,*/
	.proc		= matrix500_rtc_read_proc,
};

/*
 * Initialize and install RTC driver
 */
static int matrix500_rtc_probe(struct device *dev)
{
	int ret;
	/*struct rtc_time tm;*/

	pr_debug("%s()\n", __FUNCTION__);

        /*spin_lock(&rtc_lock);
	ret = rs372a_do_command(0, rs372a_GET_DATE, tm);
	spin_unlock(&rtc_lock);
	if(ret != 0) return ret;*/
 
	ret = register_rtc(&rtc_ops);
	if (ret) {
		printk(KERN_ERR "matrix500_rtc: could not register RTC.\n");
		
		return ret;
	}

	printk(KERN_INFO "Matrix500 RTC driver , Version : %s\n",VERSION);

	return 0;
}

/*
 * Disable and remove the RTC driver
 */
static int matrix500_rtc_remove(struct device *dev)
{
	pr_debug("%s()\n", __FUNCTION__);

	
	rtc_status = 0;

	unregister_rtc(&rtc_ops);

	return 0;
}


#define matrix500_rtc_suspend NULL
#define matrix500_rtc_resume  NULL


static struct platform_device matrix500_rtc_device = {
	.name		= DRIVER_NAME,
	.id		= -1,
};

static struct device_driver matrix500_rtc_driver = {
	.name		= DRIVER_NAME,
	.bus		= &platform_bus_type,
	.probe		= matrix500_rtc_probe,
	.remove		= matrix500_rtc_remove,
	.suspend	= matrix500_rtc_suspend,
	.resume		= matrix500_rtc_resume,
};

static int __init matrix500_rtc_init(void)
{
	int result;

	result = driver_register(&matrix500_rtc_driver);
	if (result < 0) {
		printk(KERN_ERR "Matrix500 RTC: Failed to register driver\n");
		return result;
	}

	result = platform_device_register(&matrix500_rtc_device);
	if (result < 0) {
		printk(KERN_ERR "Matrix500 RTC: Failed to register device\n");
		return result;
	}
	
	return result;
}

static void __exit matrix500_rtc_exit(void)
{
	platform_device_unregister(&matrix500_rtc_device);
	driver_unregister(&matrix500_rtc_driver);

	printk(KERN_INFO "Matrix500 RTC removed\n");
}


module_init(matrix500_rtc_init);
module_exit(matrix500_rtc_exit);

MODULE_AUTHOR("Alan Lu");
MODULE_DESCRIPTION("RTC driver for Artila Matrix500");
MODULE_LICENSE("GPL");

