/*
 *  linux/drivers/i2c/chips/rs372a.c
 *
 *  Copyright (C) 2005 James Chapman <jchapman@katalix.com>
 *
 *	based on linux/drivers/acorn/char/pcf8583.c
 *  Copyright (C) 2000 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Driver for Ricoh Semiconductor RS372A real time clock chip
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/string.h>
#include <linux/rtc.h>		/* get the user-level API */
#include <linux/bcd.h>
#include <linux/list.h>

/* Device registers */
#define RS372A_REG_SEC		(0x00)//<<4)
#define RS372A_REG_MIN		(0x01)//<<4)
#define RS372A_REG_HOUR		(0x02)//<<4)
#define RS372A_REG_DAY		(0x03)//<<4)
#define RS372A_REG_DATE		(0x04)//<<4)
#define RS372A_REG_MONTH	(0x05)//<<4)
#define RS372A_REG_YEAR		(0x06)//<<4)
#define RS372A_REG_TIMING	(0x07)//<<4)
#define RS372A_REG_ALARMA_MIN	(0x08<<4)
#define RS372A_REG_ALARMA_HR	(0x09<<4)
#define RS372A_REG_ALARMA_DAY   (0x0A<<4)
#define RS372A_REG_ALARMB_MIN	(0x0B<<4)
#define RS372A_REG_ALARMB_HR	(0x0C<<4)
#define RS372A_REG_ALARMB_DAY	(0x0D<<4)
#define RS372A_REG_CTL1		(0x0E<<4)
#define RS372A_REG_CTL2		(0x0F<<4)
/* FIXME - how do we export these interface constants? */
#define RS372A_GET_DATE		0
#define RS372A_SET_DATE		1
#define RS372A_GET_SRAM		2
#define RS372A_SET_SRAM		3
/*
 * Functions declaration
 */
static unsigned short normal_i2c[] = { 0x32, I2C_CLIENT_END };

I2C_CLIENT_INSMOD_1(rs372a);

static int rs372a_attach_adapter(struct i2c_adapter *adapter);
static int rs372a_detect(struct i2c_adapter *adapter, int address, int kind);
static void rs372a_init_client(struct i2c_client *client);
static int rs372a_detach_client(struct i2c_client *client);
static int rs372a_command(struct i2c_client *client, unsigned int cmd,
			  void *arg);

/*
 * Driver data (common to all clients)
 */
static struct i2c_driver rs372a_driver = {
	.owner		= THIS_MODULE,
	.name		= "rs372a",
	.flags		= I2C_DF_NOTIFY,
	.attach_adapter	= rs372a_attach_adapter,
	.detach_client	= rs372a_detach_client,
	.command	= rs372a_command,
};

/*
 * Client data (each client gets its own)
 */
struct rs372a_data {
	struct i2c_client client;
	struct list_head list;
};

/* Each client has this additional data */
struct rs372a_sram {
	int offset;
	int count;
	u8 *buf;
};
/*
 * Internal variables
 */
static LIST_HEAD(rs372a_clients);

static inline int rs372a_read(struct i2c_client *client, u8 reg, u8 *value)
{
	s32 tmp = i2c_smbus_read_byte_data(client, reg);

	if (tmp < 0)
		return -EIO;

	*value = tmp;

	return 0;
}

/* add by alan , 02132006 */
static inline int rs372a_write(struct i2c_client *client, u8 reg, u8 *value)
{
	s32 tmp = i2c_smbus_write_byte_data(client, reg, *value);

	if (tmp < 0)
		return -EIO;

	
	return 0;
}
/*
 * Chip access functions
 */
static int rs372a_get_datetime(struct i2c_client *client, struct rtc_time *dt)
{
	int result;
	u8 val;
	u8 offs = 0;
	u8 buf[8] = { 0 };
  
          /* this implements the 1st reading method, according
           * to the datasheet. buf[0] is initialized with
           * address ptr and transmission format register.
           */
        struct i2c_msg msg[] = {
                  { client->addr, 0, 1, buf },
                  { client->addr, I2C_M_RD, 8, buf },
        };

	if (!dt) {
		dev_dbg(&client->dev, "%s: EINVAL: dt=NULL\n", __FUNCTION__);
		return -EINVAL;
	}

	
	result = i2c_transfer(client->adapter, msg, 2);

	/*printk("%s: [%d] %02x %02x %02x %02x %02x %02x %02x\n",
		__FUNCTION__, result, buf[0], buf[1], buf[2], buf[3],
		buf[4], buf[5], buf[6]);
	*/

	if (result == 2) {
		dt->tm_sec = BCD2BIN(buf[1] & 0x7f);
		dt->tm_min = BCD2BIN(buf[2] & 0x7f);
		dt->tm_hour = BCD2BIN(buf[3] & 0x3f);
		dt->tm_wday = BCD2BIN(buf[4] & 0x07) - 1;
		dt->tm_mday = BCD2BIN(buf[5] & 0x3f);
		dt->tm_mon = BCD2BIN(buf[6] & 0x1f);//- 1;
		dt->tm_year = BCD2BIN(buf[7]);
		dt->tm_year += (dt->tm_year >= 70)?1900:2000;
		//if (buf[5] & 0x80)
		
		//	dt->tm_year += 100;

		/*printk("%s: secs=%d, mins=%d, "
			"hours=%d, mday=%d, mon=%d, year=%d, wday=%d\n",
			__FUNCTION__, dt->tm_sec, dt->tm_min,
			dt->tm_hour, dt->tm_mday,
			dt->tm_mon, dt->tm_year, dt->tm_wday);
*/
		return 0;
	}

	//dev_err(&client->dev, "error reading data! %d\n", result);
	return -EIO;
}

static int rs372a_set_datetime(struct i2c_client *client, struct rtc_time *dt)
{
	int result;
	u8 buf[8];
	u8 val;
	struct i2c_msg msg[1];

	if (!dt) {
		dev_dbg(&client->dev, "%s: EINVAL: dt=NULL\n", __FUNCTION__);
		return -EINVAL;
	}

	/*printk("%s: secs=%d, mins=%d, hours=%d, "
		"mday=%d, mon=%d, year=%d, wday=%d\n", __FUNCTION__,
		dt->tm_sec, dt->tm_min, dt->tm_hour,
		dt->tm_mday, dt->tm_mon, dt->tm_year, dt->tm_wday);
*/
	buf[0] = 0;		/* reg offset */
	buf[1] = BIN2BCD(dt->tm_sec);
	buf[2] = BIN2BCD(dt->tm_min);
	buf[3] = BIN2BCD(dt->tm_hour);
	buf[4] = BIN2BCD(dt->tm_wday) + 1;
	buf[5] = BIN2BCD(dt->tm_mday);
	buf[6] = BIN2BCD(dt->tm_mon);// + 1;
	val = dt->tm_year-1900;//-(dt->tm_year >= 70)?1900:2000;
	if (val >= 100) {
		val -= 100;
	//	buf[6] |= (1 << 7);
	}
	buf[7] = BIN2BCD(val);

	msg[0].addr = client->addr;
	msg[0].flags = 0;
	msg[0].len = sizeof(buf);
	msg[0].buf = &buf[0];

	result = i2c_transfer(client->adapter, msg, 1);
	if (result == 1)
		return 0;

	//dev_err(&client->dev, "error writing data! %d\n", result);
	return -EIO;
}

static int rs372a_command(struct i2c_client *client, unsigned int cmd,
			  void *arg)
{
	
	dev_dbg(&client->dev, "%s: cmd=%d\n", __FUNCTION__, cmd);

	switch (cmd) {
	case RS372A_GET_DATE:
		return rs372a_get_datetime(client, arg);

	case RS372A_SET_DATE:
		return rs372a_set_datetime(client, arg);
	
	default:
		return -EINVAL;
	}
}

/*
 * Public API for access to specific device. Useful for low-level
 * RTC access from kernel code.
 */
int rs372a_do_command(int bus, int cmd, void *arg)
{
	struct list_head *walk;
	struct list_head *tmp;
	struct rs372a_data *data;

	list_for_each_safe(walk, tmp, &rs372a_clients) {
		data = list_entry(walk, struct rs372a_data, list);
		if (data->client.adapter->nr == bus)
			return rs372a_command(&data->client, cmd, arg);
	}

	return -ENODEV;
}

static int rs372a_attach_adapter(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter, &addr_data, rs372a_detect);
}

/*
 * The following function does more than just detection. If detection
 * succeeds, it also registers the new chip.
 */
static int rs372a_detect(struct i2c_adapter *adapter, int address, int kind)
{
	struct i2c_client *new_client;
	struct rs372a_data *data;
	int err = 0;
	const char *name = "";

	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA |
				     I2C_FUNC_I2C))
		goto exit;

	if (!(data = kmalloc(sizeof(struct rs372a_data), GFP_KERNEL))) {
		err = -ENOMEM;
		goto exit;
	}
	memset(data, 0, sizeof(struct rs372a_data));
	INIT_LIST_HEAD(&data->list);

	/* The common I2C client data is placed right before the
	 * RS372A-specific data. 
	 */
	new_client = &data->client;
	i2c_set_clientdata(new_client, data);
	new_client->addr = address;
	new_client->adapter = adapter;
	new_client->driver = &rs372a_driver;
	new_client->flags = 0;

	/*
	 * Now we do the remaining detection. A negative kind means that
	 * the driver was loaded with no force parameter (default), so we
	 * must both detect and identify the chip. A zero kind means that
	 * the driver was loaded with the force parameter, the detection
	 * step shall be skipped. A positive kind means that the driver
	 * was loaded with the force parameter and a given kind of chip is
	 * requested, so both the detection and the identification steps
	 * are skipped.
	 *
	 * For detection, we read registers that are most likely to cause
	 * detection failure, i.e. those that have more bits with fixed
	 * or reserved values.
	 */

	/* Default to an RS372A if forced */
	if (kind == 0)
		kind = rs372a;

	if (kind < 0) {		/* detection and identification */
		u8 buf[8]={0},data, ctl2;
		struct i2c_msg msg[] = {
                  { new_client->addr, 0, 1, buf },
                  { new_client->addr, I2C_M_RD, 8, buf },
        	};
		if(i2c_transfer(new_client->adapter, msg, 2)!= 2)
			goto exit_free;

		/* Check for a valid day register value */
		ctl2 = buf[0];
		data = buf[4];
		if ((data & 0xf8))
			goto exit_free;
		
		/* Check for a valid date register value */
		data = buf[5];
		if ((data & 0xc0))
			goto exit_free;

		/* Check for a valid month register value */
		data = buf[6];
		if ((data & 0xe0))
			goto exit_free;

		msg[0].addr = new_client->addr;
		msg[0].flags = 0;
		msg[0].len = 2;
		msg[0].buf = &buf[0];
		buf[0] = 0xf0;
		buf[1] = ctl2 | 0x20; // 24h time
		i2c_transfer(new_client->adapter, msg, 1);		
		kind = rs372a;
		printk("RTC found.\n"); //alan
	}

	if (kind == rs372a)
		name = "rs372a";

	/* We can fill in the remaining client fields */
	strlcpy(new_client->name, name, I2C_NAME_SIZE);

	/* Tell the I2C layer a new client has arrived */
	if ((err = i2c_attach_client(new_client)))
		goto exit_free;

	/* Initialize the RS372A chip */
	rs372a_init_client(new_client);

	/* Add client to local list */
	list_add(&data->list, &rs372a_clients);

	return 0;

exit_free:
	kfree(data);
exit:
	return err;
}

static void rs372a_init_client(struct i2c_client *client)
{
	s32 val;

	/* Ensure that device is set in 24-hour mode */
	val = i2c_smbus_read_byte_data(client, RS372A_REG_HOUR);
	if ((val >= 0) && (val & (1 << 6)))
		i2c_smbus_write_byte_data(client, RS372A_REG_HOUR,
					  val & 0x3f);
}

static int rs372a_detach_client(struct i2c_client *client)
{
	int err;
	struct rs372a_data *data = i2c_get_clientdata(client);

	if ((err = i2c_detach_client(client)))
		return err;

	list_del(&data->list);
	kfree(data);
	return 0;
}

static int __init rs372a_init(void)
{
	return i2c_add_driver(&rs372a_driver);
}

static void __exit rs372a_exit(void)
{
	i2c_del_driver(&rs372a_driver);
}

MODULE_AUTHOR("Alan Lu");
MODULE_DESCRIPTION("RS372A RTC driver");
MODULE_LICENSE("GPL");

EXPORT_SYMBOL_GPL(rs372a_do_command);

module_init(rs372a_init);
module_exit(rs372a_exit);
