/*
 *  linux/drivers/mmc/at91rm9200_mci.c - ATMEL AT91RM9200 MCI Driver
 *
 *  Copyright (C) 2005 Cougar Creek Computing Devices Ltd, All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/*
   This is the AT91RM9200 MCI driver that has been tested with both MMC cards
   and SD-cards.  Boards that support write protect are now supported.
   The CCAT91SBC001 board does not support SD cards.

   The three entry points are at91rm9200_mci_request, at91rm9200_mci_set_ios
   and at91rm9200_mci_get_ro.

   SET IOS
     This configures the device to put it into the correct mode and clock speed
     required.

   MCI REQUEST
     MCI request processes the commands sent in the mmc_request structure. This
     can consist of a processing command and a stop command in the case of
     multiple block transfers.

     There are three main types of request, commands, reads and writes.

     Commands are straight forward. The command is submitted to the controller and
     the request function returns. When the controller generates an interrupt to indicate
     the command is finished, the response to the command are read and the mmc_request_done
     function called to end the request.

     Reads and writes work in a similar manner to normal commands but involve the PDC (DMA)
     controller to manage the transfers.

     A read is done from the controller directly to the scatterlist passed in from the request.
     Due to a bug in the controller, when a read is completed, all the words are byte
     swapped in the scatterlist buffers.

     The sequence of read interrupts is: ENDRX, RXBUFF, CMDRDY

     A write is slightly different in that the bytes to write are read from the scatterlist
     into a dma memory buffer (this is in case the source buffer should be read only). The
     entire write buffer is then done from this single dma memory buffer.

     The sequence of write interrupts is: ENDTX, TXBUFE, NOTBUSY, CMDRDY

   GET RO
     Gets the status of the write protect pin, if available.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/dma-mapping.h>

#include <linux/mmc/host.h>
#include <linux/mmc/protocol.h>

#include <linux/timer.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mach/mmc.h>
#include <asm/arch/board.h>
#include <asm/arch/AT91RM9200_MCI.h>
#include <asm/arch/pio.h> //alan

#define DRIVER_NAME "at91rm9200_mci"
//#if 1
#ifdef CONFIG_MMC_DEBUG
#define DBG(fmt...)	\
	printk(fmt)
#else
#define DBG(fmt...)	do { } while (0)
#endif

/* MCI Controller device */
static AT91PS_MCI controller = (AT91PS_MCI)AT91C_VA_BASE_MCI;
#define AT91_MCI_ERRORS	(AT91C_MCI_RINDE | AT91C_MCI_RDIRE | AT91C_MCI_RCRCE	\
		| AT91C_MCI_RENDE | AT91C_MCI_RTOE | AT91C_MCI_DCRCE		\
		| AT91C_MCI_DTOE | AT91C_MCI_OVRE | AT91C_MCI_UNRE)


#define FL_SENT_COMMAND (1 << 0)
#define FL_SENT_STOP (1 << 1)

/* Low level type for this driver
 */
struct at91mci_host
{
	struct mmc_host *mmc;
	struct mmc_command *cmd;
	struct mmc_request *request;

	struct at91_mmc_data *board;
	int present;

	/*
	 * Flag indicating when the command has been sent. This is used to
	 * work out whether or not to send the stop
	 */
	unsigned int flags;

	/* DMA buffer used for transmitting */
	unsigned int* buffer;
	dma_addr_t physical_address;
	unsigned int total_length;

	/* Latest in the scatterlist that has been enabled for transfer, but not freed */
	int in_use_index;

	/* Latest in the scatterlist that has been enabled for transfer */
	int transfer_index;

	/* Timer for timeouts */
	struct timer_list timer;
};
/*
 * Enable the controller
 */
static void at91rm9200_mci_enable(void)
{
	AT91_SYS->PMC_PCER = 1 << AT91C_ID_MCI;		/* Enable the peripheral clock */
	controller->MCI_CR = AT91C_MCI_MCIEN;
	controller->MCI_IDR = 0xFFFFFFFF;
	controller->MCI_DTOR = AT91C_MCI_DTOMUL_1048576 | AT91C_MCI_DTOCYC;
	controller->MCI_MR = 0x834A;
	controller->MCI_SDCR = 0x0;
}

/*
 * Disable the controller
 */
static void at91rm9200_mci_disable(void)
{
	controller->MCI_CR = AT91C_MCI_MCIDIS | 0x80;
	AT91_SYS->PMC_PCDR = 1 << AT91C_ID_MCI;		/* Disable the peripheral clock */
}

/*
 * Reset the controller and restore most of the state
 */
static void at91_reset_host(struct at91mci_host *host)
{
	unsigned long flags;
	u32 mr;
	u32 sdcr;
	u32 dtor;
	u32 imr;
	
	local_irq_save(flags);
	imr = controller->MCI_IMR;

	controller->MCI_IDR= 0xffffffff;

	/* save current state */
	mr = controller->MCI_MR & 0x7fff;
	sdcr = controller->MCI_SDCR;
	dtor = controller->MCI_DTOR;

	/* reset the controller */
	controller->MCI_CR = AT91C_MCI_MCIDIS | AT91C_MCI_SWRST;
	//AT91_SYS->PMC_PCDR = 1 << AT91C_ID_MCI;		/* Disable the peripheral clock */
	udelay(100);
	//AT91_SYS->PMC_PCER = 1 << AT91C_ID_MCI;		/* Disable the peripheral clock */
	/* restore state */
	controller->MCI_CR=AT91C_MCI_MCIEN;
	controller->MCI_MR=mr;
	controller->MCI_SDCR= sdcr;
	controller->MCI_DTOR= dtor;
	controller->MCI_IER= imr;

	/* make sure sdio interrupts will fire */
	imr = controller->MCI_SR;

	local_irq_restore(flags);
}

static void at91_timeout_timer(unsigned long data)
{
	struct at91mci_host *host;

	host = (struct at91mci_host *)data;

	if (host->request) {
		printk("Timeout waiting end of packet\n");

		if (host->cmd && host->cmd->data) {
			host->cmd->data->error = -ETIMEDOUT;
		} else {
			if (host->cmd)
				host->cmd->error = -ETIMEDOUT;
			else
				host->request->cmd->error = -ETIMEDOUT;
		}
		//AT91_SYS->PMC_PCDR = 1 << AT91C_ID_MCI;		/* Disable the peripheral clock */
		//udelay(200);
		//AT91_SYS->PMC_PCER = 1 << AT91C_ID_MCI;		/* Disable the peripheral clock */
		//at91rm9200_mci_disable();
		//at91rm9200_mci_enable();
		at91_reset_host(host);
		mmc_request_done(host->mmc, host->request);
	}
}

/*
 * Copy from sg to a dma block - used for transfers
 */
static inline void at91mci_sg_to_dma(struct at91mci_host *host, struct mmc_data *data)
{
	unsigned int len, i, size;
	unsigned *dmabuf = host->buffer;

	size = data->blksz * data->blocks;
	len = data->sg_len;

	/*
	 * Just loop through all entries. Size might not
	 * be the entire list though so make sure that
	 * we do not transfer too much.
	 */
	for (i = 0; i < len; i++) {
		struct scatterlist *sg;
		int amount;
		int index;
		unsigned int *sgbuffer;

		sg = &data->sg[i];

		sgbuffer = kmap_atomic(sg->page, KM_BIO_SRC_IRQ) + sg->offset;
		amount = min(size, sg->length);
		size -= amount;
		amount /= 4;

		for (index = 0; index < amount; index++) {
			*dmabuf++ = swab32(sgbuffer[index]);
		}
		kunmap_atomic(sgbuffer, KM_BIO_SRC_IRQ);

		if (size == 0)
			break;
	}

	/*
	 * Check that we didn't get a request to transfer
	 * more data than can fit into the SG list.
	 */
	BUG_ON(size != 0);
}

/*
 * Prepare a dma read
 */
static void at91mci_pre_dma_read(struct at91mci_host *host)
{
	int i;
	struct scatterlist *sg;
	struct mmc_command *cmd;
	struct mmc_data *data;

	DBG("pre dma read\n");

	cmd = host->cmd;
	if (!cmd) {
		DBG("no command\n");
		return;
	}

	data = cmd->data;
	if (!data) {
		DBG("no data\n");
		return;
	}

	for (i = 0; i < 2; i++) {
		/* nothing left to transfer */
		if (host->transfer_index >= data->sg_len) {
			DBG("Nothing left to transfer (index = %d)\n", host->transfer_index);
			break;
		}

		/* Check to see if this needs filling */
		if (i == 0) {
			if (controller->MCI_RCR != 0) {
				DBG("Transfer active in current\n");
				continue;
			}
		}
		else {
			if (controller->MCI_RNCR != 0) {
				DBG("Transfer active in next\n");
				continue;
			}
		}

		/* Setup the next transfer */
		DBG("Using transfer index %d\n", host->transfer_index);

		sg = &data->sg[host->transfer_index++];
		DBG("sg = %p\n", sg);

		sg->dma_address =
			dma_map_page(NULL, sg->page, sg->offset, sg->length, DMA_FROM_DEVICE);

		DBG("dma address = %08X, length = %d\n", sg->dma_address, sg->length);

		if (i == 0) {
			controller->MCI_RPR = sg->dma_address;
			controller->MCI_RCR = sg->length / 4;
		}
		else {
			controller->MCI_RNPR = sg->dma_address;
			controller->MCI_RNCR = sg->length / 4;
		}
	}

	DBG("pre dma read done\n");
}

/*
 * Handle after a dma read
 */
static void at91mci_post_dma_read(struct at91mci_host *host)
{
	struct mmc_command *cmd;
	struct mmc_data *data;

	DBG("post dma read\n");

	cmd = host->cmd;
	if (!cmd) {
		DBG("no command\n");
		return;
	}

	data = cmd->data;
	if (!data) {
		DBG("no data\n");
		return;
	}

	while (host->in_use_index < host->transfer_index) {
		unsigned int *buffer;
		int index;
		int len;

		struct scatterlist *sg;

		DBG("finishing index %d\n", host->in_use_index);

		sg = &data->sg[host->in_use_index++];

		DBG("Unmapping page %08X\n", sg->dma_address);

		dma_unmap_page(NULL, sg->dma_address, sg->length, DMA_FROM_DEVICE);

		/* Swap the contents of the buffer */
		buffer = kmap_atomic(sg->page, KM_BIO_SRC_IRQ) + sg->offset;
		DBG("buffer = %p, length = %d\n", buffer, sg->length);

		len = sg->length / 4;

		for (index = 0; index < len; index++) {
			buffer[index] = swab32(buffer[index]);
		}
		kunmap_atomic(buffer, KM_BIO_SRC_IRQ);
		flush_dcache_page(sg->page);

		data->bytes_xfered += sg->length;
	}

	/* Is there another transfer to trigger? */
	if (host->transfer_index < data->sg_len)
		at91mci_pre_dma_read(host);
	else {
		controller->MCI_IDR = AT91C_MCI_ENDRX;
		controller->MCI_IER = AT91C_MCI_RXBUFF;
		//controller->MCI_IER = AT91C_MCI_RXBUFF;
		//controller->MCI_PTCR = AT91C_PDC_RXTDIS | AT91C_PDC_TXTDIS;
	}

	DBG("post dma read done\n");
}

/*
 * Handle transmitted data
 */
static void at91rm9200_mci_handle_transmitted(struct at91mci_host *host)
{
	struct mmc_command *cmd;
	struct mmc_data *data;

	DBG("Handling the transmit\n");

	/* Disable the transfer */
	controller->MCI_PTCR = AT91C_PDC_RXTDIS | AT91C_PDC_TXTDIS;

	/* Now wait for cmd ready */
	controller->MCI_IDR = AT91C_MCI_TXBUFE;
	//

	cmd = host->cmd;
	if (!cmd) return;

	data = cmd->data;
	if (!data) return;

	if (cmd->data->blocks > 1) {
		DBG("multiple write : wait for BLKE...\n");
		controller->MCI_IER = AT91C_MCI_BLKE;
	} else
		controller->MCI_IER = AT91C_MCI_NOTBUSY;
}

/*
 * Update bytes tranfered count during a write operation
 */
static void at91_mci_update_bytes_xfered(struct at91mci_host *host)
{
	struct mmc_data *data;

	/* always deal with the effective request (and not the current cmd) */

	if (host->request->cmd && host->request->cmd->error != 0)
		return;

	if (host->request->data) {
		data = host->request->data;
		if (data->flags & MMC_DATA_WRITE) {
			/* card is in IDLE mode now */
			DBG("-> bytes_xfered %d, total_length = %d\n",
				data->bytes_xfered, host->total_length);
			data->bytes_xfered = data->blksz * data->blocks;
		}
	}
}


/*Handle after command sent ready*/
static int at91_mci_handle_cmdrdy(struct at91mci_host *host)
{
	if (!host->cmd)
		return 1;
	else if (!host->cmd->data) {
		if (host->flags & FL_SENT_STOP) {
			/*After multi block write, we must wait for NOTBUSY*/
			controller->MCI_IER =  AT91C_MCI_NOTBUSY;
		} else return 1;
	} else if (host->cmd->data->flags & MMC_DATA_WRITE) {
		/*After sendding multi-block-write command, start DMA transfer*/
		controller->MCI_IER =  AT91C_MCI_TXBUFE | AT91C_MCI_BLKE;
		controller->MCI_PTCR=  AT91C_PDC_TXTEN ;
	}

	/* command not completed, have to wait */
	return 0;
}

/*
 * Send a command
 * return the interrupts to enable
 */
static unsigned int at91rm9200_mci_send_command(struct at91mci_host *host, struct mmc_command *cmd)
{
	unsigned int cmdr;
	unsigned int block_length;
	struct mmc_data *data = cmd->data;
	//struct mmc_host *mmc = host->mmc; //alan

	unsigned int blocks;
	unsigned int ier = 0;

	host->cmd = cmd;

	/* Not sure if this is needed */
#if 1 //alan : 0 -> 1
	if ((controller->MCI_SR & AT91C_MCI_RTOE) && (cmd->opcode == 1)) {
		DBG("Clearing timeout\n");
		controller->MCI_ARGR = 0;
		controller->MCI_CMDR = AT91C_MCI_OPDCMD;
		while (!(controller->MCI_SR & AT91C_MCI_CMDRDY)) {
			/* spin */
			//DBG("Clearing: SR = %08X\n", at91_mci_read(host, AT91_MCI_SR));
		}
	}
#endif
	cmdr = cmd->opcode;

	if (mmc_resp_type(cmd) == MMC_RSP_NONE)
		cmdr |= AT91C_MCI_RSPTYP_NO;
	else {
		/* if a response is expected then allow maximum response latancy */
		cmdr |= AT91C_MCI_MAXLAT;
		/* set 136 bit response for R2, 48 bit response otherwise */
		if (mmc_resp_type(cmd) == MMC_RSP_R2)
			cmdr |= AT91C_MCI_RSPTYP_136;
		else
			cmdr |= AT91C_MCI_RSPTYP_48;
	}

	if (data) {
		block_length =  data->blksz;
		blocks = data->blocks;

		/* always set data start - also set direction flag for read */
		if (data->flags & MMC_DATA_READ)
			cmdr |= (AT91C_MCI_TRDIR | AT91C_MCI_TRCMD_START);
		else if (data->flags & MMC_DATA_WRITE)
			cmdr |= AT91C_MCI_TRCMD_START;

		if (data->flags & MMC_DATA_STREAM)
			cmdr |= AT91C_MCI_TRTYP_STREAM;
		if (data->blocks > 1)
			cmdr |= AT91C_MCI_TRTYP_MULTIPLE;
	}
	else {
		block_length = 0;
		blocks = 0;
	}
	if (host->flags & FL_SENT_STOP)
		cmdr |= AT91C_MCI_TRCMD_STOP;
	
	/*if (mmc->ios.bus_mode == MMC_BUSMODE_OPENDRAIN) {
		cmdr |= AT91C_MCI_OPDCMD;
	}*/ //alan

	/*
	 * Set the arguments and send the command
	 */
	DBG("Sending command %d as %08X, arg = %08X, blocks = %d, length = %d (MR = %08X)\n",
		cmd->opcode, cmdr, cmd->arg, blocks, block_length, controller->MCI_MR);

	if (!data) {
		
		controller->MCI_PTCR = (AT91C_PDC_TXTDIS | AT91C_PDC_RXTDIS);
		controller->MCI_RPR  = 0;
		controller->MCI_RCR  = 0;
		controller->MCI_RNPR  = 0;
		controller->MCI_RNCR  = 0;
		controller->MCI_TPR  = 0;
		controller->MCI_TCR  = 0;
		controller->MCI_TNPR  = 0;
		controller->MCI_TNCR  = 0;

		controller->MCI_ARGR = cmd->arg;
		controller->MCI_CMDR = cmdr;
		return AT91C_MCI_CMDRDY;
	}else {

	controller->MCI_MR &= ~0xFFFF8000;	/* zero block length and PDC mode */
	controller->MCI_MR |= (block_length << 16) | AT91C_MCI_PDCMODE;

	/*
	 * Disable the PDC controller
	 */
	controller->MCI_PTCR = AT91C_PDC_RXTDIS | AT91C_PDC_TXTDIS;

	if (cmdr & AT91C_MCI_TRCMD_START) {
		data->bytes_xfered = 0;
		host->transfer_index = 0;
		host->in_use_index = 0;
		if (cmdr & AT91C_MCI_TRDIR) {
			/*
			 * Handle a read
			 */
			host->buffer = NULL;
			host->total_length = 0;

			at91mci_pre_dma_read(host);
			ier = AT91C_MCI_ENDRX /* | AT91C_MCI_RXBUFF */;
			//ier = AT91C_MCI_RXBUFF;
		} else {
			/*
			 * Handle a write
			 */
			host->total_length = block_length * blocks;
				host->buffer = kmalloc(host->total_length, GFP_KERNEL);
				if (!host->buffer) {
					DBG("Can't alloc tx buffer\n");
					cmd->error = -ENOMEM;
					mmc_request_done(host->mmc, host->request);
					return -1;
				}

			at91mci_sg_to_dma(host, data);

				host->physical_address = dma_map_single(NULL,
						host->buffer, host->total_length,
						DMA_TO_DEVICE);

			controller->MCI_TPR = host->physical_address;
			controller->MCI_TCR = host->total_length / 4;
			ier = AT91C_MCI_TXBUFE;
			}
		}
	}

	/*
	 * Send the command and then enable the PDC - not the other way round as
	 * the data sheet says
	 */

	controller->MCI_ARGR = cmd->arg;
	controller->MCI_CMDR = cmdr;

	if (cmdr & AT91C_MCI_TRCMD_START) {
		if (cmdr & AT91C_MCI_TRDIR) {
			controller->MCI_PTCR = AT91C_PDC_RXTEN;
		}
		else {
			controller->MCI_PTCR = AT91C_PDC_TXTEN;
		}
	}

	return ier;
}

/*
 * Wait for a command to complete
 */
static void at91mci_process_command(struct at91mci_host *host, struct mmc_command *cmd)
{
	unsigned int ier;
	
	ier = at91rm9200_mci_send_command(host, cmd);

	DBG("setting ier to %08X\n", ier);

	/* Stop on errors or the required value */
	controller->MCI_IER =  AT91_MCI_ERRORS | ier;
}

/*
 * Process the next step in the request
 */
static void at91mci_process_next(struct at91mci_host *host)
{
	
	if (!(host->flags & FL_SENT_COMMAND)) {
		host->flags |= FL_SENT_COMMAND;
		at91mci_process_command(host, host->request->cmd);
	}
	else if ((!(host->flags & FL_SENT_STOP)) && host->request->stop) {
		host->flags |= FL_SENT_STOP;
		at91mci_process_command(host, host->request->stop);;
	}
	else {
		del_timer(&host->timer);
		at91_reset_host(host);
		mmc_request_done(host->mmc, host->request);
	}
}

/*
 * Handle a command that has been completed
 */
static void at91mci_completed_command(struct at91mci_host *host, unsigned int status)
{
	struct mmc_command *cmd = host->cmd;
	//struct mmc_data *data = cmd->data;
	//int i;
	
	controller->MCI_IDR = 0xffffffff;

	cmd->resp[0] = controller->MCI_RSPR[0];
	cmd->resp[1] = controller->MCI_RSPR[1];
	cmd->resp[2] = controller->MCI_RSPR[2];
	cmd->resp[3] = controller->MCI_RSPR[3];

	if (host->buffer) {
		dma_unmap_single(NULL,
				host->physical_address, host->total_length,
				DMA_TO_DEVICE);
		kfree(host->buffer);
		host->buffer = NULL;
	}

	DBG("Status = %08X/%08x [%08X %08X %08X %08X]\n",
		 status, controller->MCI_SR,
		 cmd->resp[0], cmd->resp[1], cmd->resp[2], cmd->resp[3]);

	if (status & AT91_MCI_ERRORS) {
#if 1
		if ( (status & AT91C_MCI_RCRCE) &&
			((cmd->opcode == MMC_SEND_OP_COND) ||
				 (cmd->opcode == SD_APP_OP_COND))) {
			cmd->error = MMC_ERR_NONE;
		}
		else {
			if (status & (AT91C_MCI_RTOE | AT91C_MCI_DTOE)) {
				cmd->error = MMC_ERR_TIMEOUT;
			}
			else if (status & (AT91C_MCI_RCRCE | AT91C_MCI_DCRCE)) {
				cmd->error = MMC_ERR_BADCRC;
			}
			else if (status & (AT91C_MCI_OVRE | AT91C_MCI_UNRE)) {
				cmd->error = MMC_ERR_FIFO;
			}
			else {
				cmd->error = MMC_ERR_FAILED;
			}
#endif

			DBG("Error detected and set to %d (cmd = %d, retries = %d)\n",
				 cmd->error, cmd->opcode, cmd->retries);
		}
	}
	else {
		cmd->error = 0;//MMC_ERR_NONE;
	}

	at91mci_process_next(host);
}

/*
 * Handle an MMC request
 */
static void at91rm9200_mci_request(struct mmc_host *mmc, struct mmc_request *mrq)
{
	struct at91mci_host *host = mmc_priv(mmc);
	host->request = mrq;
	host->flags = 0;

	mod_timer(&host->timer, jiffies +  HZ);

	at91mci_process_next(host);
}

/*
 * Set the IOS
 */
static void at91rm9200_mci_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	int clkdiv;
	struct at91mci_host *host = mmc_priv(mmc);

	DBG("Clock %uHz, busmode %u, powermode %u, Vdd %u\n",
		ios->clock, ios->bus_mode, ios->power_mode, ios->vdd);

	if (ios->clock == 0) {
		/* Disable the MCI controller */
		controller->MCI_CR = AT91C_MCI_MCIDIS;
		clkdiv = 0;
	}
	else {
		/* Enable the MCI controller */
		controller->MCI_CR = AT91C_MCI_MCIEN;

		if ( (at91_master_clock % (ios->clock * 2)) == 0 )
			clkdiv = ((at91_master_clock / ios->clock) / 2) - 1;
		else
			clkdiv = (at91_master_clock / ios->clock) / 2;

		DBG("clkdiv = %d. mcck = %ld\n", clkdiv,
			at91_master_clock / (2 * (clkdiv + 1)));
	}
	if (ios->bus_width == MMC_BUS_WIDTH_4 && host->board->wire4) {
		DBG("Setting controller bus width to 4\n");
		controller->MCI_SDCR |= AT91C_MCI_SCDBUS;
	}
	else {
		DBG("Setting controller bus width to 1\n");
		controller->MCI_SDCR &= ~AT91C_MCI_SCDBUS;
		//controller->MCI_SDCR |= AT91C_MCI_SCDBUS;
	}

	/* Set the clock divider */
	controller->MCI_MR = (controller->MCI_MR & ~0x000000FF) | clkdiv;

	/* maybe switch power to the card */
	if (host->board->vcc_pin) {
		switch (ios->power_mode) {
		case MMC_POWER_OFF:
			at91_set_gpio_output(host->board->vcc_pin, 0);
			break;
		case MMC_POWER_UP:
			at91_set_gpio_output(host->board->vcc_pin, 1);
			break;
		case MMC_POWER_ON:
				break;
			default:
				WARN_ON(1);
		}
	}
}
//static DEFINE_SPINLOCK(mmc_lock);
/*
 * Handle an interrupt
 */
static irqreturn_t at91rm9200_mci_irq(int irq, void *devid, struct pt_regs *regs)
{
	struct at91mci_host *host = devid;
	int completed = 0;
	unsigned int int_status, int_mask;
//spin_lock(&mmc_lock);
	int_status = controller->MCI_SR;
	int_mask = controller->MCI_IMR;

	DBG("MCI irq: status = %08X, %08X, %08X\n", int_status, controller->MCI_IMR,
		int_status & controller->MCI_IMR);

	int_status = int_status & int_mask;

	if (int_status & AT91_MCI_ERRORS) {
		completed = 1;

		if (int_status & AT91C_MCI_UNRE)
			DBG("MMC: Underrun error\n");
		if (int_status & AT91C_MCI_OVRE)
			DBG("MMC: Overrun error\n");
		if (int_status & AT91C_MCI_DTOE)
			DBG("MMC: Data timeout\n");
		if (int_status & AT91C_MCI_DCRCE)
			DBG("MMC: CRC error in data\n");
		if (int_status & AT91C_MCI_RTOE)
			DBG("MMC: Response timeout\n");
		if (int_status & AT91C_MCI_RENDE)
			DBG("MMC: Response end bit error\n");
		if (int_status & AT91C_MCI_RCRCE)
			DBG("MMC: Response CRC error\n");
		if (int_status & AT91C_MCI_RDIRE)
			DBG("MMC: Response direction error\n");
		if (int_status & AT91C_MCI_RINDE)
			DBG("MMC: Response index error\n");
	} else {
		if (int_status & AT91C_MCI_TXBUFE) {
			DBG("TX buffer empty\n");
			at91rm9200_mci_handle_transmitted(host);
		}
	
		if (int_status & AT91C_MCI_RXBUFF) {
			//printk("RX buffer full\n");
			DBG("RX buffer full\n");
			//int i;
			//controller->MCI_IER = AT91C_MCI_CMDRDY;
			controller->MCI_PTCR = AT91C_PDC_RXTDIS | AT91C_PDC_TXTDIS;
			controller->MCI_IDR  = AT91C_MCI_RXBUFF | AT91C_MCI_ENDRX;
			completed = 1;
		}

		if (int_status & AT91C_MCI_ENDTX) {
			DBG("Transmit has ended\n");
		}

		if (int_status & AT91C_MCI_ENDRX) {
			DBG("Receive has ended\n");
			at91mci_post_dma_read(host);
		}

		if (int_status & AT91C_MCI_NOTBUSY) {
			DBG("Card is ready\n");
			at91_mci_update_bytes_xfered(host);
			completed = 1;
		}

		if (int_status & AT91C_MCI_DTIP) {
			DBG("Data transfer in progress\n");
		}

		if (int_status & AT91C_MCI_BLKE) {
			DBG("Black transfer has ended\n");
			if (host->request->data && host->request->data->blocks > 1) {
				/* multi block write : complete multi write
				 * command and send stop */
				completed = 1;
			} else {
				controller->MCI_IER = AT91C_MCI_NOTBUSY;
			}
		}

		if (int_status & AT91C_MCI_TXRDY) {
			DBG("Ready to transmit\n");
		}

		if (int_status & AT91C_MCI_RXRDY) {
			DBG("Ready to receive\n");
		}

		if (int_status & AT91C_MCI_CMDRDY) {
			DBG("Command ready\n");
		 	completed = at91_mci_handle_cmdrdy(host);
		}
	}
	//controller->MCI_IDR = int_status;

	if (completed) {
		DBG("Completed command\n");
		controller->MCI_IDR = 0xffffffff;
		at91mci_completed_command(host, int_status);
	} else {
		controller->MCI_IDR = int_status;
	}
	return IRQ_HANDLED;
}



static irqreturn_t
at91rm9200_mmc_det_irq(int irq, void *_host, struct pt_regs *regs)
{
	struct at91mci_host *host = _host;
	int present = !at91_get_gpio_value(irq);

	/* we expect this irq on both insert and remove,
	 * and use a short delay to debounce.
	 */
	
	if (present != host->present) {
		host->present = present;
		printk("%s: card %s\n", mmc_hostname(host->mmc),
			present ? "insert" : "remove"); //alan : DBG -> printk
		if (!present) {
			DBG("****** Resetting SD-card bus width ******\n");
			controller->MCI_SDCR = 0;
		}
		#if 0 // micro sd
		mmc_detect_change(host->mmc, present == 1 ?msecs_to_jiffies(2000):msecs_to_jiffies(100));
		#else
		mmc_detect_change(host->mmc, msecs_to_jiffies(100));
		#endif
	}
	return IRQ_HANDLED;
}



int at91rm9200_mci_get_ro(struct mmc_host *mmc)
{
/*	int read_only = 0;
	struct at91mci_host *host = mmc_priv(mmc);
	if( host->board->wp_pin ) {
		read_only = at91_get_gpio_value(host->board->wp_pin);
		printk(KERN_WARNING "%s: card is %s\n",
					mmc_hostname(mmc),
					(read_only ? "read-only" : "read-write") );
	}
	else {
		printk(KERN_WARNING "%s: host does not "
				"support reading read-only "
				"switch.  Assuming write-enable.\n",
				mmc_hostname(mmc));
	}
	return read_only;*/
#ifdef CONFIG_MMC_NO_PROTECT
	
	return 0;
#else
	return at91_get_gpio_value(AT91_PIN_PD20); //alan
#endif
}



static struct mmc_host_ops at91rm9200_mci_ops = {
	.request	= at91rm9200_mci_request,
	.set_ios	= at91rm9200_mci_set_ios,
	.get_ro		= at91rm9200_mci_get_ro,
};

/*
 * Probe for the device
 */
static int at91rm9200_mci_probe(struct device *dev)
{
	struct mmc_host *mmc;
	struct at91mci_host *host;
	int ret;

	DBG("Probe MCI devices\n");

	mmc = mmc_alloc_host(sizeof(struct at91mci_host), dev);
	if (!mmc) {
		DBG("Failed to allocate mmc host\n");

		ret = -ENOMEM;
		return ret;
	}

	mmc->ops = &at91rm9200_mci_ops;
	mmc->f_min = 375000;
	mmc->f_max = 25000000;
	mmc->ocr_avail = MMC_VDD_32_33 | MMC_VDD_33_34;

	host = mmc_priv(mmc);
	host->mmc = mmc;
	host->buffer = NULL;
	host->board = dev->platform_data;
	if (host->board->wire4)
#ifdef CONFIG_MMC_SUPPORT_4WIRE
		mmc->caps |= MMC_CAP_4_BIT_DATA;
#else
		printk("AT91 MMC: 4 wire bus mode not supported by this driver - using 1 wire\n");
#endif
		
	/*
	 * Allocate the MCI interrupt
	 */
	ret = request_irq(AT91C_ID_MCI, at91rm9200_mci_irq, SA_SHIRQ, DRIVER_NAME, host);
	if (ret) {
		DBG("Failed to request MCI interrupt\n");
		return ret;
	}
	setup_timer(&host->timer, at91_timeout_timer, (unsigned long)host);
	dev_set_drvdata(dev, mmc);
	host->present = !at91_get_gpio_value(AT91_PIN_PD19);//alan
	/*
	 * Add host to MMC layer
	 */
	//printk("mmc: det_pin->%d\n", host->board->det_pin);
	if (host->board->det_pin)
		host->present = !at91_get_gpio_value(host->board->det_pin);
	else
		host->present = -1;
	mmc_add_host(mmc);

	/*
	 * monitor card insertion/removal if we can
	 */
	if (host->board->det_pin) {//alan
		ret = request_irq(host->board->det_pin, at91rm9200_mmc_det_irq,
				SA_SAMPLE_RANDOM, DRIVER_NAME, host);
		if (ret)
			DBG("couldn't allocate MMC detect irq\n");
	}//alan

	DBG(KERN_INFO "Added MCI driver\n");

	return 0;
}

/*
 * Remove a device
 */
static int at91rm9200_mci_remove(struct device *dev)
{
	struct mmc_host *mmc = dev_get_drvdata(dev);
	struct at91mci_host *host;

	if (!mmc) return -1;

	host = mmc_priv(mmc);

	if (host->present != -1) {
		free_irq(host->board->det_pin, host);
		cancel_delayed_work(&host->mmc->detect);
	}
	/* alan */
	//free_irq(AT91C_ID_MCI, host);
	mmc_free_host(mmc);

	DBG("Removed\n");

	return 0;
}

static struct device_driver at91rm9200_mci_driver = {
	.name		= DRIVER_NAME,
	.bus		= &platform_bus_type,
	.probe		= at91rm9200_mci_probe,
	.remove		= at91rm9200_mci_remove,
};

/* Initialise the MCI controller
 */
static int __init at91rm9200_mci_init(void)
{
	int result;

#ifdef CONFIG_MMC_NO_PROTECT
	at91_set_gpio_value(AT91_PIN_PD20, 0);
#endif
	AT91_CfgPIO_MCISlotA();//alan

	at91rm9200_mci_disable();
	at91rm9200_mci_enable();

	result = driver_register(&at91rm9200_mci_driver);
	if (result < 0) {
		printk(KERN_ERR "AT91 MCI: Failed to register driver\n");
		return result;
	}

	//result = platform_device_register(&at91rm9200_mci_device);
	printk(KERN_INFO "AT91RM9200 MCI initialized\n");

	return 0;
}

/*
 * Release the controller
 */
static void __exit at91rm9200_mci_exit(void)
{
	at91rm9200_mci_disable();
	driver_unregister(&at91rm9200_mci_driver);

	printk(KERN_INFO "AT91RM9200 MCI removed\n");
}

module_init(at91rm9200_mci_init);
module_exit(at91rm9200_mci_exit);

MODULE_DESCRIPTION("AT91RM9200 Multimedia Card Interface driver");
MODULE_AUTHOR("Nick Randell");
MODULE_LICENSE("GPL");
