/*
 * $Id: artila-map.c,v 1.13 2004/11/04 13:24:14 gleixner Exp $
 *
 * Handle mapping of the NOR flash on M508 boards
 *
 * Copyright 2002 SYSGO Real-Time Solutions GmbH
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define M508_ADDR0 0x10000000      /* physical properties of flash */
#define M508_SIZE0 0x01000000
#define M508_ADDR1 0x30000000      /* physical properties of flash */
#define M508_SIZE1 0x01000000

#ifdef CONFIG_MTD_RAM
#define M508_ADDR2 0x40000000      /* physical properties of flash */
#define M508_SIZE2 0x00080000
#define NUM_FLASHBANKS 3
#else
#define NUM_FLASHBANKS 2
#endif 

#define BUSWIDTH     2

/* can be { "cfi_probe", "jedec_probe", "map_rom", NULL } */
#define PROBETYPES { "jedec_probe", NULL }

#define MSG_PREFIX "Artila:"   /* prefix for our printk()'s */
#define MTDID      "artila-%d"  /* for mtdparts= partitioning */

static struct mtd_info *artila_mtd[NUM_FLASHBANKS];

int sram_flag;

static struct map_info artila_map[NUM_FLASHBANKS] = {
	{
		.name = "phys_mapped_flash",
		.size = M508_SIZE0,
		.bankwidth = BUSWIDTH,
	},
	{
		.name = "phys_mapped_flash#2",
		.size = M508_SIZE1,
		.bankwidth = BUSWIDTH,
	},
#ifdef CONFIG_MTD_RAM
	{
		.name = "sram",
		.size = M508_SIZE2,
		.bankwidth = BUSWIDTH,
		
	},
#endif 
};

#ifdef CONFIG_MTD_PARTITIONS

/*
 * MTD partitioning stuff 
 */
static struct mtd_partition static_partitions[] =
{
	{
		.name = "FileSystem",
		.size = 0x1000000,
		.offset = 0x00000000
	},
};
#ifdef CONFIG_MTD_RAM
static struct mtd_partition static_partitions2[] =
{
	{
		.name = "sram",
		.size = 0x80000,
		.offset = 0x00000000,
	},
};
#endif
static int mtd_parts_nb[NUM_FLASHBANKS];
static struct mtd_partition *mtd_parts[NUM_FLASHBANKS];

#endif

static const char *probes[] = { "cmdlinepart", NULL };

int __init init_artila(void)
{
	//static const char *rom_probe_types[] = PROBETYPES;
	static const char *rom_probe_types[] = { "cfi_probe", "jedec_probe", NULL };
	const char **type;
	const char *part_type = 0;
	int i , flashbanks;
	static struct { u_long addr; u_long size; } pt[NUM_FLASHBANKS] = {
	  { M508_ADDR0, M508_SIZE0 },
	  { M508_ADDR1, M508_SIZE1 },
#ifdef CONFIG_MTD_RAM
	  { M508_ADDR2, M508_SIZE2 },
#endif
        };
	int devicesfound = 0;
	
	if(sram_flag)
		flashbanks = 3;
	else
		flashbanks = 2;


	//for(i=0; i<NUM_FLASHBANKS; i++)
	for(i=0; i<flashbanks; i++)
	{
		printk(KERN_NOTICE MSG_PREFIX "probing 0x%08lx at 0x%08lx\n",
		       pt[i].size, pt[i].addr);

		artila_map[i].phys = pt[i].addr;
		artila_map[i].virt = ioremap(pt[i].addr, pt[i].size);
		if (!artila_map[i].virt) {
			printk(KERN_NOTICE MSG_PREFIX "failed to ioremap\n");
			return -EIO;
		}

		simple_map_init(&artila_map[i]);

		artila_mtd[i] = 0;
		if(artila_map[i].name[0] != 's') { //ace change 20100226
			type = rom_probe_types;
			for(; !artila_mtd[i] && *type; type++) {
				artila_mtd[i] = do_map_probe(*type, &artila_map[i]);
			}
		}else
			artila_mtd[i] = do_map_probe("map_ram", &artila_map[i]);

		if (artila_mtd[i]) {
			artila_mtd[i]->owner = THIS_MODULE;
			devicesfound++;
#ifdef CONFIG_MTD_PARTITIONS
			mtd_parts_nb[i] = parse_mtd_partitions(artila_mtd[i], 
							       probes,
							       &mtd_parts[i], 
							       0);
			//printk("\nmtd_parts_nb[%d] = %d\n\n",i,mtd_parts_nb[i]);
			if (mtd_parts_nb[i] > 0) {
				part_type = "command line";
			} else {
#ifdef CONFIG_MTD_RAM
				if(i == 1){
#endif
				mtd_parts[i] = static_partitions;
				mtd_parts_nb[i] = ARRAY_SIZE(static_partitions); 
#ifdef CONFIG_MTD_RAM
				}else {
				mtd_parts[i] = static_partitions2;
				mtd_parts_nb[i] = ARRAY_SIZE(static_partitions2);
				artila_mtd[i]->erasesize= 0x8000;//32k
				}
#endif
				part_type = "static";
			}

			printk(KERN_NOTICE MSG_PREFIX
			       "using %s partition definition\n", 
			       part_type);
			//mdelay(1000);
			add_mtd_partitions(artila_mtd[i], 
					   mtd_parts[i], mtd_parts_nb[i]);
#else
			add_mtd_device(artila_mtd[i]);

#endif
		}
		else 
			iounmap((void *)artila_map[i].virt);
	}
	return devicesfound == 0 ? -ENXIO : 0;
}

static void __exit cleanup_artila(void)
{
	int i;
	for (i=0; i<NUM_FLASHBANKS; i++) {
		if (artila_mtd[i]) {
#ifdef CONFIG_MTD_PARTITIONS
			del_mtd_partitions(artila_mtd[i]);
#else
			del_mtd_device(artila_mtd[i]);
#endif
			map_destroy(artila_mtd[i]);
			iounmap((void *)artila_map[i].virt);
			artila_map[i].virt = 0;
		}
	}
}

static int __init init_sram(char *str){
	sram_flag = 1;
	return 0;
}
__setup("sram=", init_sram);

module_init(init_artila);
module_exit(cleanup_artila);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Ace Yang");
MODULE_DESCRIPTION("MTD map driver for Artila Intel P30/33");
