/*
 *  Driver for TI TLV320AIC2XTS on Artila M501 soundcard
 *  Copyright (C) 2002 Tomas Kasparek <tomas.kasparek@seznam.cz>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License.
 * 
 * History:
 *
 * 2002-03-13   Tomas Kasparek  initial release - based on h3600-uda1341.c from OSS
 * 2002-03-20   Tomas Kasparek  playback over ALSA is working
 * 2002-03-28   Tomas Kasparek  playback over OSS emulation is working
 * 2002-03-29   Tomas Kasparek  basic capture is working (native ALSA)
 * 2002-03-29   Tomas Kasparek  capture is working (OSS emulation)
 * 2002-04-04   Tomas Kasparek  better rates handling (allow non-standard rates)
 * 2003-02-14   Brian Avery     fixed full duplex mode, other updates
 * 2003-02-20   Tomas Kasparek  merged updates by Brian (except HAL)
 * 2003-04-19   Jaroslav Kysela recoded DMA stuff to follow 2.4.18rmk3-hh24 kernel
 *                              working suspend and resume
 * 2003-04-28   Tomas Kasparek  updated work by Jaroslav to compile it under 2.5.x again
 *                              merged HAL layer (patches from Brian)
 */

/* $Id: sa11xx-uda1341.c,v 1.23 2005/09/09 13:22:34 tiwai Exp $ */

/***************************************************************************************************
*
* To understand what Alsa Drivers should be doing look at "Writing an Alsa Driver" by Takashi Iwai
* available in the Alsa doc section on the website		
* 
* A few notes to make things clearer. The TLV320AIC2X is hooked up to Serial port 4 on the SA1100.
* We are using  SSP mode to talk to the TLV320AIC2X. The TLV320AIC2X bit & wordselect clocks are generated
* by this UART. Unfortunately, the clock only runs if the transmit buffer has something in it.
* So, if we are just recording, we feed the transmit DMA stream a bunch of 0x0000 so that the
* transmit buffer is full and the clock keeps going. The zeroes come from FLUSH_BASE_PHYS which
* is a mem loc that always decodes to 0's w/ no off chip access.
*
* Some alsa terminology:
*	frame => num_channels * sample_size  e.g stereo 16 bit is 2 * 16 = 32 bytes
*	period => the least number of bytes that will generate an interrupt e.g. we have a 1024 byte
*             buffer and 4 periods in the runtime structure this means we'll get an int every 256
*             bytes or 4 times per buffer.
*             A number of the sizes are in frames rather than bytes, use frames_to_bytes and
*             bytes_to_frames to convert.  The easiest way to tell the units is to look at the
*             type i.e. runtime-> buffer_size is in frames and its type is snd_pcm_uframes_t
*             
*	Notes about the pointer fxn:
*	The pointer fxn needs to return the offset into the dma buffer in frames.
*	Interrupts must be blocked before calling the dma_get_pos fxn to avoid race with interrupts.
*
*	Notes about pause/resume
*	Implementing this would be complicated so it's skipped.  The problem case is:
*	A full duplex connection is going, then play is paused. At this point you need to start xmitting
*	0's to keep the record active which means you cant just freeze the dma and resume it later you'd
*	need to	save off the dma info, and restore it properly on a resume.  Yeach!
*
*	Notes about transfer methods:
*	The async write calls fail.  I probably need to implement something else to support them?
* 
***************************************************************************************************/

#include <linux/config.h>
#include <sound/driver.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <asm/arch-at91rm9200/artila.h>
#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#include <asm/hardware.h>
#include <asm/arch/AT91RM9200_SSC.h>
#include <asm/mach-types.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>


#undef DEBUG_MODE
#undef DEBUG_FUNCTION_NAMES
#define USB_MODE 0
#define AUDIO_MCLK		12000000
#define MCLK			60000000
#define AUDIO_RATE_DEFAULT	44100
#define BITS_BY_SLOT  16
#define SLOT_BY_FRAME  2
/* TCMR */
/*#define I2S_ASY_MASTER_TX_SETTING(nb_bit_by_slot, nb_slot_by_frame) ( \
	((nb_bit_by_slot * nb_slot_by_frame / 2 - 1) << 24) | \
	AT91C_SSC_START_FALL_RF | \
	AT91C_SSC_CKS_TK |\
	AT91C_SSC_CKO_NONE | \
	((0<<16) & AT91C_SSC_STTDLY))
*/
#define I2S_ASY_MASTER_TX_SETTING(nb_bit_by_slot, nb_slot_by_frame) ( \
	((nb_bit_by_slot * nb_slot_by_frame / 2 - 1) << 24) | \
	AT91C_SSC_START_FALL_RF | \
	AT91C_SSC_CKO_CONTINOUS | \
	AT91C_SSC_CKS_DIV |\
	((0<<16) & AT91C_SSC_STTDLY))

  //AT91C_SSC_CKS_DIV |\

/* TFMR */
/*#define I2S_ASY_TX_FRAME_SETTING(nb_bit_by_slot, nb_slot_by_frame) ( \
	AT91C_SSC_FSOS_LOW | \
	AT91C_SSC_MSBF | \
	(((nb_bit_by_slot - 1) << 16) & AT91C_SSC_FSLEN) | \
	(((nb_slot_by_frame - 1) << 8) & AT91C_SSC_DATNB) | \
	(nb_bit_by_slot - 1))*/
#define I2S_ASY_TX_FRAME_SETTING(nb_bit_by_slot, nb_slot_by_frame) ( \
	AT91C_SSC_FSOS_NEGATIVE | \
	AT91C_SSC_MSBF | \
	(((nb_bit_by_slot - 1) << 16) & AT91C_SSC_FSLEN) | \
	(((nb_slot_by_frame - 1) << 8) & AT91C_SSC_DATNB) | \
	(nb_bit_by_slot - 1))

/*#define I2S_ASY_TX_FRAME_SETTING(nb_bit_by_slot, nb_slot_by_frame) ( \
	| AT91C_SSC_FSEDGE | \
	AT91C_SSC_FSOS_NEGATIVE | \
	AT91C_SSC_MSBF | \
	(((nb_bit_by_slot - 1) << 16) & AT91C_SSC_FSLEN) | \
	(((nb_slot_by_frame - 1) << 8) & AT91C_SSC_DATNB) | \
	(nb_bit_by_slot - 1))*/

/* {{{ Type definitions */

MODULE_AUTHOR("Alan Lu");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Artila M501 TLV320AIC2XTS driver for ALSA");
MODULE_SUPPORTED_DEVICE("{{TLV320AIC2X}}");

static char *id = NULL;	/* ID for this card */

module_param(id, charp, 0444);
MODULE_PARM_DESC(id, "ID string for M501 TLV320AIC2X soundcard.");

typedef struct audio_stream {
	char *id;		/* identification string */
	int stream_id;		/* numeric identification */	
	int active;		/* we are using this stream for transfer now */
	int period;		/* current transfer period */
	int periods;		/* current count of periods registerd in the DMA engine */
	int tx_spin;		/* are we recoding - flag used to do DMA trans. for sync */
	unsigned int old_offset;
	spinlock_t dma_lock;	/* for locking in DMA operations (see dma-sa1100.c in the kernel) */
	dma_addr_t	dma_ptr;
	dma_addr_t	dma_addr;
	size_t		period_size;
	size_t		period_bytes;
	size_t		buffer_size;
	size_t		dma_bytes;
	size_t		dma_xfer;
	snd_pcm_substream_t *stream;
}audio_stream_t;

typedef struct snd_card_at91_tlv320 {
	snd_card_t *card;
	struct l3_client *uda1341;
	snd_pcm_t *pcm;
	long samplerate;
	audio_stream_t s[2];	/* playback & capture */
	AT91PS_SSC pSSC;
	int irq;
} at91_tlv320_t;

static struct snd_card_at91_tlv320 *at91_tlv320 = NULL;

static unsigned int rates[] = {
	8000,  10666, 10985, 14647,
	16000, 21970, 22050, 24000,
	29400, 32000, 44100, 48000,
};

static snd_pcm_hw_constraint_list_t hw_constraints_rates = {
	.count	= ARRAY_SIZE(rates),
	.list	= rates,
	.mask	= 0,
};


static void AT91_PDC_Close(AT91PS_SSC pSSC){
	/* close PDC */
	pSSC->SSC_PTCR = AT91C_SSC_RXDIS | AT91C_SSC_TXDIS;
	pSSC->SSC_TNPR = 0;
	pSSC->SSC_TNCR = 0;
	pSSC->SSC_RNPR = 0;
	pSSC->SSC_RNCR = 0;
	pSSC->SSC_TPR  = 0;
	pSSC->SSC_TCR  = 0;
	pSSC->SSC_RPR  = 0;
	pSSC->SSC_RCR  = 0; 
}

/************************ I2C section ***********************/

#define I2C_DRIVERID_TLV320 (0xFDCB)

static struct i2c_client * tlv320_client = NULL;

static int tlv320_attach_adapter(struct i2c_adapter *adapter);
static int tlv320_detect_client(struct i2c_adapter *adapter, int address);
static int tlv320_detach_client(struct i2c_client *client);
static int tlv320_init_client(struct i2c_client * new_client);

static int tlv320_initialized = 0;
static int cur_left_vol;
static int cur_right_vol;
static unsigned char tlv320_reg[]={
					0x97, /*0: Left-in volume */ 	
					0x97, /*1: Right-in volume */
 					0xf0, //0xe0, /*2: Left-headphone volume */
					0xf0,//e0, /*3: Right-headphone volume */
					0x12, /*4: Analog Audio Control */ 
					0x04, /*5: Digital Aduio Control */
					0x47, /*6: Power Down Control */
					0x02, /*7: Digital Format */
					0x20, /*8: Sample Rate Control */
					0x01  /*9: Digital Activation */
				};
static struct i2c_driver tlv320_driver = {
        name:           "TLV320AIC2X driver",
        id:             I2C_DRIVERID_TLV320,
        flags:          I2C_DF_NOTIFY,
        attach_adapter: &tlv320_attach_adapter,
        detach_client:  &tlv320_detach_client,
        command:        NULL,
};

struct tlv320_data {
        int arf; /* place holder for furture use */
};

int tlv320_cleanup(void) {
        if (!tlv320_initialized)
                return -ENODEV;
	i2c_smbus_write_byte_data(tlv320_client, 9<<1, 0);
	i2c_smbus_write_byte_data(tlv320_client, 5<<1, 0x08);
	i2c_smbus_write_byte_data(tlv320_client, 6<<1, 0xff);
        i2c_del_driver(&tlv320_driver);
        tlv320_initialized = 0;
	
        return 0;
}

int tlv320_init(void) {
        int rc; 

        if (tlv320_initialized)
                return 0;

        printk("tlv320 I2C driver version 1.0\n");

        if ((rc = i2c_add_driver(&tlv320_driver))) {
                printk("tlv320: Driver registration failed, module not inserted.\n");
                tlv320_cleanup();
                return rc;
        }
        tlv320_initialized = 1;

        return 0;
}

static int tlv320_attach_adapter(struct i2c_adapter *adapter) {
        tlv320_detect_client(adapter, 0x1b); //0x40 for AIC20
        return 0;
}

static int tlv320_detect_client(struct i2c_adapter *adapter, int address) {
        int rc = 0;
        struct i2c_client *new_client;

////        struct tlv320_data *data;
        const char *client_name = "DAC TLV320 Digital Equalizer";

        new_client = kmalloc(
                        sizeof(struct i2c_client) + sizeof(struct tlv320_data),
                        GFP_KERNEL);
        if (!new_client) {
                rc = -ENOMEM;
                goto bail;
        }

        new_client->addr = address;
        new_client->adapter = adapter;
        new_client->driver = &tlv320_driver;
        new_client->flags = 0;

        strcpy(new_client->name,client_name);

////        new_client->id = tlv320_id++; /* Automatically unique */

        if (tlv320_init_client(new_client)) {
                rc = -ENODEV;
                goto bail;
        }

        /* Tell the i2c layer a new client has arrived */
        if (i2c_attach_client(new_client)) {
                rc = -ENODEV;
                goto bail;
        }
bail:
        if (rc && new_client)
                kfree(new_client);
        return rc;
}

static int tlv320_detach_client(struct i2c_client *client) {
        if (client == tlv320_client)
                tlv320_client = NULL;

        i2c_detach_client(client);
        kfree(client);

        return 0;
}

/* volume is in percents. 0 == mute, 100 == max */
int tlv320_set_volume(uint left_vol, uint right_vol) {
        unsigned short voldata;

        if (!tlv320_client)
                return -1;

	if (left_vol  > 100) left_vol  = 100;
	if (right_vol > 100) right_vol = 100;

        cur_left_vol = left_vol;
        cur_right_vol = right_vol;

	/* max volume is 0x38 OR MAYBE 0x3f */
	left_vol  = 0x3f * left_vol  / 100;
	right_vol = 0x3f * right_vol / 100;

        voldata = (left_vol  << 8) | right_vol;
	//printk("%x\n",voldata);i2c_smbus_write_byte_data
        if (i2c_smbus_write_word_data(tlv320_client, 2, voldata) < 0) {
                printk("daca: failed to set volume \n");
                return -1;
        }

        return 0;
}

void tlv320_get_volume(uint *vol) {
        *vol =  (cur_left_vol << 8) | cur_right_vol;
	//printk("VOL = %d\n",vol);
}
//EXPORT_SYMBOL(tlv320_get_volume);
//EXPORT_SYMBOL(tlv320_set_volume);

struct i2c_client * tlv_get_i2c_clinet(void) {
	return tlv320_client;
}
EXPORT_SYMBOL(tlv_get_i2c_clinet);

static int tlv320_init_client(struct i2c_client * new_client) {
        int i=0;	
	
	/* initialize TLV320AIC2X */
	i2c_smbus_write_byte_data(new_client, 0x0f<<1, 0);
	mdelay(500);
	for(i=0; i<10; i++)
		i2c_smbus_write_byte_data(new_client, i<<1, tlv320_reg[i]);

	tlv320_client = new_client;

	
        return 0;
}
/* {{{ Clock and sample rate stuff */



static void at91_tlv320_set_audio_clock(long val)
{
	switch (val) {
	case 24000: case 32000: case 48000:	/* 00: 12.288 MHz */
		
		break;

	case 22050: case 29400: case 44100:	/* 01: 11.2896 MHz */
		
		break;

	case 8000: case 10666: case 16000:	/* 10: 4.096 MHz */
		
		break;

	case 10985: case 14647: case 21970:	/* 11: 5.6245 MHz */
		
		break;
	}
}

static void at91_tlv320_set_samplerate(at91_tlv320_t *at91_tlv320, long rate)
{
	//int clk_div = 0;
	//int clk=0;
	unsigned int baud_value;
	/* We don't want to mess with clocks when frames are in flight */
	
	/* wait for any frame to complete */
	//udelay(125);

	/*
	 * We have the following clock sources:
	 * 4.096 MHz, 5.6245 MHz, 11.2896 MHz, 12.288 MHz
	 * Those can be divided either by 256, 384 or 512.
	 * This makes up 12 combinations for the following samplerates...
	 */
	if (rate >= 48000)
		rate = 48000;
	else if (rate >= 44100)
		rate = 44100;
	else if (rate >= 32000)
		rate = 32000;
	else if (rate >= 29400)
		rate = 29400;
	else if (rate >= 24000)
		rate = 24000;
	else if (rate >= 22050)
		rate = 22050;
	else if (rate >= 21970)
		rate = 21970;
	else if (rate >= 16000)
		rate = 16000;
	else if (rate >= 14647)
		rate = 14647;
	else if (rate >= 10985)
		rate = 10985;
	else if (rate >= 10666)
		rate = 10666;
	else
		rate = 8000;

	/* Set the external clock generator */

	at91_tlv320_set_audio_clock(rate);

	/* Select the clock divisor */
	if(AUDIO_MCLK == 12000000){
		switch (rate) {
			case 48000:
				//snd_printk("set 12MHz, 48000\n");
				
				if(USB_MODE)
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x01);
				else
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x02);
				
				i2c_smbus_write_byte_data(tlv320_client, 5<<1, 0x06);
				baud_value = 20;
			break;
			case 44100:
				//snd_printk("set 12MHz, 44100\n");
				
				if(USB_MODE)
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x023);	
				else
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x20);
				
				i2c_smbus_write_byte_data(tlv320_client, 5<<1, 0x04);
				baud_value = 21;
			break;
			case 32000:
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x19);
				baud_value = 30;
			break;
			case 22050:
				//snd_printk("set 12MHz, 44100\n");
				i2c_smbus_write_byte_data(tlv320_client, 5<<1, 0x04);
				if(USB_MODE)
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x023);	
				else
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x28);
				i2c_smbus_write_byte_data(tlv320_client, 5<<1, 0x04);
				baud_value = 43;
			break;
			case 8000:
				i2c_smbus_write_byte_data(tlv320_client, 8<<1, 0x0d);
				baud_value = 117;
			break;
		}
	}

	/*if (rate == 0)
           baud_value = 0;
        else
        {
           //baud_value = (unsigned int) (MCLK * 10)/(2*rate);
		baud_value = (unsigned int)(MCLK)/(rate*SLOT_BY_FRAME*BITS_BY_SLOT*2);
           if ((baud_value % 10) >= 5)
                  baud_value = (baud_value / 10) + 1;
           else
                  baud_value /= 10;
        }*/
	//if(rate == 48000) baud_value = 20;
	//else if(rate == 44100) baud_value = 21;
	//snd_printk("set Rate:%ld, CMR:%d\n", rate, baud_value);
        at91_tlv320->pSSC->SSC_CMR = baud_value;
	
	at91_tlv320->samplerate = rate;
}

/* }}} */

static irqreturn_t at91ssc_interrupt(int irq, void *dev_id, struct pt_regs *regs){
	at91_tlv320_t *chip = (at91_tlv320_t*) dev_id;
	audio_stream_t *s = &chip->s[SNDRV_PCM_STREAM_PLAYBACK]; 
	snd_pcm_substream_t * substream = s->stream;
	snd_pcm_runtime_t *runtime = substream->runtime;
	AT91PS_SSC pSSC = chip->pSSC;
	int status =  (pSSC->SSC_SR & pSSC->SSC_IMR); 
	//dma_addr_t dma_addr;
	//size_t dma_bytes;
	
	if(!s->active || !s->tx_spin){
		snd_printk("inactive interrupt!\n"); 
		return IRQ_HANDLED;
	}
	
	spin_lock(&s->dma_lock);
	s->period++;
	s->periods++;
	//dma_bytes = frames_to_bytes(runtime,runtime->period_size);
	
	//dma_addr = runtime->dma_addr + (dma_bytes * s->period);
	s->dma_ptr += s->period_bytes;
	if((s->dma_ptr - s->dma_addr) >= s->dma_bytes){
		s->period=0;
		//dma_addr = runtime->dma_addr;
		s->dma_ptr = s->dma_addr;
	}
	/*spin_unlock(&s->dma_lock);
	snd_pcm_period_elapsed(substream);
	spin_lock(&s->dma_lock);*/
	if(status & AT91C_SSC_TXBUFE){
		snd_printk("AT91C_SSC_TXBUFE\n");
		pSSC->SSC_PTCR = AT91C_PDC_TXTDIS;
		pSSC->SSC_TPR  = s->dma_ptr;
		pSSC->SSC_TCR  = s->period_bytes/s->dma_xfer;
		pSSC->SSC_PTCR = AT91C_PDC_TXTEN;
	}
	if(status & AT91C_SSC_ENDTX){
		//snd_printk("AT91C_SSC_ENDTX:0x%x\n", dma_addr);
		pSSC->SSC_TNPR = s->dma_ptr;
		pSSC->SSC_TNCR = s->period_bytes/s->dma_xfer;	
		
	}
	spin_unlock(&s->dma_lock);
	//snd_printk("dma_ptr:0x%x, bytes:%d\n", s->dma_ptr, dma_bytes);
	snd_pcm_period_elapsed(substream);
	return IRQ_HANDLED;	
}
/* {{{ HW init and shutdown */

static void at91_tlv320_audio_init(at91_tlv320_t *at91_tlv320)
{
	unsigned long flags;
	unsigned int ssc_pio_lines = ((unsigned int) AT91C_PB12_TF2) |
	                             ((unsigned int) AT91C_PB13_TK2) |
	                             ((unsigned int) AT91C_PB14_TD2);
	AT91PS_SSC pSSC;
	
	
	at91_tlv320->s[SNDRV_PCM_STREAM_PLAYBACK].id = "TLV320AIC2X out";
	at91_tlv320->s[SNDRV_PCM_STREAM_PLAYBACK].stream_id = SNDRV_PCM_STREAM_PLAYBACK;
	
	at91_tlv320->s[SNDRV_PCM_STREAM_CAPTURE].id = "TLV320AIC2X in";
	at91_tlv320->s[SNDRV_PCM_STREAM_CAPTURE].stream_id = SNDRV_PCM_STREAM_CAPTURE;
	

	/* Initialize the TLV320AIC2X internal state */
       
	/* Setup the uarts */
	local_irq_save(flags);
	pSSC = at91_tlv320->pSSC = (AT91PS_SSC) AT91C_VA_BASE_SSC2;
	at91_tlv320->irq = AT91C_ID_SSC2;
	AT91_SYS->PIOB_PDR = ssc_pio_lines;
	AT91_SYS->PIOB_ODR = ssc_pio_lines;
	AT91_SYS->PIOB_IFDR = ssc_pio_lines;
	AT91_SYS->PIOB_CODR = ssc_pio_lines;
	AT91_SYS->PIOB_IDR = ssc_pio_lines;
	AT91_SYS->PIOB_MDDR = ssc_pio_lines;
	AT91_SYS->PIOB_PPUDR = ssc_pio_lines;
	AT91_SYS->PIOB_ASR = ssc_pio_lines;
	AT91_SYS->PIOB_OWDR = ssc_pio_lines;
	AT91_SYS->PIOD_ASR = AT91C_PD7_PCK0;  
	AT91_SYS->PIOD_PDR = AT91C_PD7_PCK0;
	//AT91_SYS->PMC_PCKR[0] = AT91C_PMC_CSS_MAIN_CLK|AT91C_PMC_PRES_CLK;
	//AT91_SYS->CKGR_PLLAR = 0x007c0018 | AT91C_CKGR_SRCA;
	//mdelay(500);
	//snd_printk("CKGR_PLLAR:0x%x\n", AT91_SYS->CKGR_PLLAR);
	AT91_SYS->PMC_PCKR[0] = AT91C_PMC_CSS_PLLA_CLK|AT91C_PMC_PRES_CLK_16;
	AT91_SYS->PMC_SCER = AT91C_PMC_PCK0; // Enable PCK0
	AT91_PDC_Close(pSSC);	
	AT91_SYS->PMC_PCER = 1<<AT91C_ID_SSC2;	/* Enable SSC2 Clock */
	pSSC->SSC_CR = AT91C_SSC_SWRST;
	mdelay(500);
	
	//
	pSSC->SSC_TFMR = I2S_ASY_TX_FRAME_SETTING(BITS_BY_SLOT, SLOT_BY_FRAME);
	pSSC->SSC_TCMR = I2S_ASY_MASTER_TX_SETTING(BITS_BY_SLOT, SLOT_BY_FRAME);
	if(request_irq(at91_tlv320->irq, at91ssc_interrupt, 0, at91_tlv320->card->driver, at91_tlv320)<0)
		printk(KERN_ERR "at91_tlv320: Can't allocate IRQ%d\n", at91_tlv320->irq);

	tlv320_init();
	at91_tlv320_set_samplerate(at91_tlv320, at91_tlv320->samplerate);
	/* Power down TLV320 DAC */
	i2c_smbus_write_byte_data(tlv320_client, 6<<1, 0xFF);
	local_irq_restore(flags);
	
}

static void at91_tlv320_audio_shutdown(at91_tlv320_t *at91_tlv320)
{
	/* mute on */

	
	/* disable the audio power and all signals leading to the audio chip */
	

	/* power off and mute off */
	/* FIXME - is muting off necesary??? */

}

#define SET_VOLUME 0xc001
#define GET_VOLUME 0xc002

/* {{{ trigger & timer */

static int snd_at91_tlv320_trigger(snd_pcm_substream_t * substream, int cmd)
{
	at91_tlv320_t *chip = snd_pcm_substream_chip(substream);
	int stream_id = substream->pstr->stream;
	audio_stream_t *s = &chip->s[stream_id];
	audio_stream_t *s1 = &chip->s[stream_id ^ 1];
	snd_pcm_runtime_t *runtime = substream->runtime;
	AT91PS_SSC pSSC = chip->pSSC;
	int err = 0;
	//uint vol;
	/* note local interrupts are already disabled in the midlevel code */
	spin_lock(&s->dma_lock);
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		if(stream_id == SNDRV_PCM_STREAM_PLAYBACK){
			//snd_printk("trigger_start!\n");
			/* requested stream startup */
			
			s->dma_ptr = s->dma_addr;
			pSSC->SSC_TPR  = s->dma_ptr;
			pSSC->SSC_TCR  = s->period_bytes/s->dma_xfer;//bytes;
			
			s->dma_ptr += s->period_bytes;
			pSSC->SSC_TNPR = s->dma_ptr;
			pSSC->SSC_TNCR =s->period_bytes/s->dma_xfer;//bytes;
			/* Unmute DAC */
			//i2c_smbus_write_byte_data(tlv320_client, 5<<1, 0);

			pSSC->SSC_CR = AT91C_SSC_TXEN;
			pSSC->SSC_PTCR = AT91C_PDC_TXTEN;
			pSSC->SSC_IER = AT91C_SSC_ENDTX|AT91C_SSC_TXBUFE;
			s->active = 1;
			s->tx_spin = 1;
		}
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		if(stream_id == SNDRV_PCM_STREAM_PLAYBACK){
			//snd_printk("trigger_stop!\n");
			/* requested stream startup */
			s->tx_spin = 0;
			s->active = 0;
			AT91_PDC_Close(pSSC);
			pSSC->SSC_IDR = AT91C_SSC_ENDTX|AT91C_SSC_TXBUFE;
			pSSC->SSC_PTCR = AT91C_PDC_TXTDIS;
			pSSC->SSC_CR = AT91C_SSC_TXDIS;
			/* Mute DAC */
			//i2c_smbus_write_byte_data(tlv320_client, 5<<1, 0x08);
			i2c_smbus_write_byte_data(tlv320_client, 6<<1, 0xFF);
		}
		
		
		break;
	case SNDRV_PCM_TRIGGER_SUSPEND:
		s->active = 0;

		//s->old_offset = audio_get_dma_pos(s) + 1;

		s->periods = 0;
		break;
	case SNDRV_PCM_TRIGGER_RESUME:
		s->active = 1;
		s->tx_spin = 0;
		//audio_process_dma(s);
		if (stream_id == SNDRV_PCM_STREAM_CAPTURE && !s1->active) {
			s1->tx_spin = 1;
		//	audio_process_dma(s1);
		}
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:

		s->active = 0;
		if (stream_id == SNDRV_PCM_STREAM_PLAYBACK) {
			if (s1->active) {
				s->tx_spin = 1;
		//		s->old_offset = audio_get_dma_pos(s) + 1;
	
		//		audio_process_dma(s);
			}
		} else {
			if (s1->tx_spin) {
				s1->tx_spin = 0;
	
			}
		}
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
		s->active = 1;
		if (s->old_offset) {
			s->tx_spin = 0;
		//	audio_process_dma(s);
			break;
		}
		if (stream_id == SNDRV_PCM_STREAM_CAPTURE && !s1->active) {
			s1->tx_spin = 1;
		//	audio_process_dma(s1);
		}

		break;
	/*case SET_VOLUME:
		vol = *((int __user *) arg);
		tlv320_set_volume(vol,vol);
		return 0;
	break;
	case GET_VOLUME:
		//vol = *((uint __user *) argp);
		tlv320_get_volume(&vol);
		return vol;
	break;*/
	default:
		err = -EINVAL;
		break;
	}
	spin_unlock(&s->dma_lock);	
	return err;
}

static int snd_at91_tlv320_prepare(snd_pcm_substream_t * substream)
{
	at91_tlv320_t *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;
	audio_stream_t *s = &chip->s[substream->pstr->stream];
	
	s->period_size = runtime->period_size;
	s->period_bytes = frames_to_bytes(runtime,s->period_size);
	s->dma_addr	= runtime->dma_addr;
	s->dma_ptr 	= s->dma_addr;
	s->dma_bytes	= runtime->dma_bytes;
	s->buffer_size	= runtime->buffer_size;
	s->dma_xfer	= 2; //16-bit 
	//snd_printk("dma_addr:0x%x, dma_bytes:%d, period_bytes:%d\n",s->dma_addr, s->dma_bytes, s->period_bytes);
        /* set requested samplerate */
	/* Power On TLV320 */
	tlv320_init_client(tlv320_client);
		
	at91_tlv320_set_samplerate(chip, runtime->rate);
			
	s->period = 0;
	s->periods = 0;
	s->tx_spin = 0;
	s->active = 0;
        return 0;
}

static snd_pcm_uframes_t snd_at91_tlv320_pointer(snd_pcm_substream_t * substream)
{
	at91_tlv320_t *chip = snd_pcm_substream_chip(substream);
	int stream_id = substream->pstr->stream;
	audio_stream_t *s = &chip->s[stream_id];
	audio_stream_t *s1 = &chip->s[stream_id ^ 1];
	snd_pcm_runtime_t *runtime = substream->runtime;
	AT91PS_SSC pSSC = chip->pSSC;
	size_t offset;
	unsigned long flags;

	spin_lock_irqsave(&s->dma_lock, flags);
	
	offset = bytes_to_frames(runtime, pSSC->SSC_TPR - s->dma_addr);
	if(offset == s->buffer_size) offset = 0;
	//snd_printk("offset(%d frames,%d bytes)->current:%x\n", offset, (ptr-start), ptr);//pSSC->SSC_TPR);//TCR);
	spin_unlock_irqrestore(&s->dma_lock, flags);
	return offset;
}
/* }}} */

static snd_pcm_hardware_t snd_at91_tlv320_capture =
{
	.info			= (SNDRV_PCM_INFO_INTERLEAVED |
				   SNDRV_PCM_INFO_BLOCK_TRANSFER |
				   SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_MMAP_VALID |
				   SNDRV_PCM_INFO_PAUSE | SNDRV_PCM_INFO_RESUME),
	.formats		= SNDRV_PCM_FMTBIT_S16_LE,
	.rates			= (SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |\
				   SNDRV_PCM_RATE_22050 | SNDRV_PCM_RATE_32000 |\
				   SNDRV_PCM_RATE_44100 | SNDRV_PCM_RATE_48000 |\
				   SNDRV_PCM_RATE_KNOT),
	.rate_min		= 8000,
	.rate_max		= 48000,
	.channels_min		= 2,
	.channels_max		= 2,
	.buffer_bytes_max	= 64*1024,
	.period_bytes_min	= 64,
	.period_bytes_max	= 16*1024,
	.periods_min		= 2,
	.periods_max		= 255,
	.fifo_size		= 0,
};

static snd_pcm_hardware_t snd_at91_tlv320_playback =
{
	.info			= (SNDRV_PCM_INFO_INTERLEAVED |
				   SNDRV_PCM_INFO_BLOCK_TRANSFER |
				   SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_MMAP_VALID),
	.formats		= SNDRV_PCM_FMTBIT_S16_LE,
	.rates			= (SNDRV_PCM_RATE_8000 | 							   SNDRV_PCM_RATE_16000 |\
                                   SNDRV_PCM_RATE_22050 | SNDRV_PCM_RATE_32000 |\
				   SNDRV_PCM_RATE_44100 | SNDRV_PCM_RATE_48000 |\
				   SNDRV_PCM_RATE_KNOT),
	.rate_min		= 8000,
	.rate_max		= 48000,
	.channels_min		= 2,
	.channels_max		= 2,
	.buffer_bytes_max	= 64*1024,
	.period_bytes_min	= 64, 
	.period_bytes_max	= 16*1024,
	.periods_min		= 2,
	.periods_max		= 255,
	.fifo_size		= 0,
};

static int snd_card_at91_tlv320_open(snd_pcm_substream_t * substream)
{
	at91_tlv320_t *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;
	int stream_id = substream->pstr->stream;
	int err;

	chip->s[stream_id].stream = substream;

	if (stream_id == SNDRV_PCM_STREAM_PLAYBACK)
		runtime->hw = snd_at91_tlv320_playback;
	else
		runtime->hw = snd_at91_tlv320_capture;
	if ((err = snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS)) < 0)
		return err;
	if ((err = snd_pcm_hw_constraint_list(runtime, 0, SNDRV_PCM_HW_PARAM_RATE, &hw_constraints_rates)) < 0)
		return err;
        
	return 0;
}

static int snd_card_at91_tlv320_close(snd_pcm_substream_t * substream)
{
	at91_tlv320_t *chip = snd_pcm_substream_chip(substream);

	chip->s[substream->pstr->stream].stream = NULL;
	return 0;
}

/* {{{ HW params & free */

static int snd_at91_tlv320_hw_params(snd_pcm_substream_t * substream,
					snd_pcm_hw_params_t * hw_params)
{
        
	return snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params));
}

static int snd_at91_tlv320_hw_free(snd_pcm_substream_t * substream)
{
	return snd_pcm_lib_free_pages(substream);
}

/* }}} */

static snd_pcm_ops_t snd_card_at91_tlv320_playback_ops = {
	.open			= snd_card_at91_tlv320_open,
	.close			= snd_card_at91_tlv320_close,
	.ioctl			= snd_pcm_lib_ioctl,
	.hw_params	        = snd_at91_tlv320_hw_params,
	.hw_free	        = snd_at91_tlv320_hw_free,
	.prepare		= snd_at91_tlv320_prepare,
	.trigger		= snd_at91_tlv320_trigger,
	.pointer		= snd_at91_tlv320_pointer,
};

static snd_pcm_ops_t snd_card_at91_tlv320_capture_ops = {
	.open			= snd_card_at91_tlv320_open,
	.close			= snd_card_at91_tlv320_close,
	.ioctl			= snd_pcm_lib_ioctl,
	.hw_params	        = snd_at91_tlv320_hw_params,
	.hw_free	        = snd_at91_tlv320_hw_free,
	.prepare		= snd_at91_tlv320_prepare,
	.trigger		= snd_at91_tlv320_trigger,
	.pointer		= snd_at91_tlv320_pointer,
};

static int __init snd_card_at91_tlv320_pcm(at91_tlv320_t *at91_tlv320, int device)
{
	snd_pcm_t *pcm;
	int err;

	if ((err = snd_pcm_new(at91_tlv320->card, "TLV320AIC2X PCM", device, 1, 0, &pcm)) < 0) 
		return err;

	
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV, 
					      0,
					      64*1024, 64*1024);

	/*snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_CONTINUOUS, 
					      snd_dma_continuous_data(GFP_KERNEL),
					      0, 64*1024);
*/
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_card_at91_tlv320_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &snd_card_at91_tlv320_capture_ops);
	pcm->private_data = at91_tlv320;
	pcm->info_flags = 0;
	strcpy(pcm->name, "TLV320AIC2X PCM");

	at91_tlv320_audio_init(at91_tlv320);

	at91_tlv320->pcm = pcm;

	return 0;
}

/* }}} */

/* {{{ module init & exit */

#ifdef CONFIG_PM

static int snd_at91_tlv320_suspend(snd_card_t *card, pm_message_t state)
{
	at91_tlv320_t *chip = card->pm_private_data;

	snd_pcm_suspend_all(chip->pcm);

	
	at91_tlv320_audio_shutdown(chip);
	return 0;
}

static int snd_at91_tlv320_resume(snd_card_t *card)
{
	at91_tlv320_t *chip = card->pm_private_data;

	at91_tlv320_audio_init(chip);
	

	return 0;
}
#endif /* COMFIG_PM */

void snd_at91_tlv320_free(snd_card_t *card)
{
	at91_tlv320_t *chip = card->private_data;

	if(chip->irq > 0) free_irq(chip->irq, chip);
	tlv320_cleanup();
	at91_tlv320 = NULL;
	card->private_data = NULL;
	kfree(chip);
}

static int __init at91_tlv320_init(void)
{
	int err;
	snd_card_t *card;

	if (!machine_is_matrix520() && !machine_is_matrix510() && !machine_is_m501()){
		printk("This driver is only for Artila Matrix!\n");
		return -ENODEV;
	}

	/* register the soundcard */
	card = snd_card_new(-1, id, THIS_MODULE, sizeof(at91_tlv320_t));
	if (card == NULL)
		return -ENOMEM;

	at91_tlv320 = kzalloc(sizeof(*at91_tlv320), GFP_KERNEL);
	if (at91_tlv320 == NULL)
		return -ENOMEM;	
	spin_lock_init(&at91_tlv320->s[0].dma_lock);
	spin_lock_init(&at91_tlv320->s[1].dma_lock);
         
	card->private_data = (void *)at91_tlv320;
	card->private_free = snd_at91_tlv320_free;

	at91_tlv320->card = card;
	at91_tlv320->samplerate = AUDIO_RATE_DEFAULT;

	// mixer
	/*if ((err = snd_chip_uda1341_mixer_new(at91_tlv320->card, &at91_tlv320->uda1341)))
		goto nodev;
*/
	// PCM
	if ((err = snd_card_at91_tlv320_pcm(at91_tlv320, 0)) < 0)
		goto nodev;
        
	snd_card_set_generic_pm_callback(card,
				     snd_at91_tlv320_suspend, snd_sa11_uda1341_resume,
				     at91_tlv320);

	strcpy(card->driver, "TLV320AIC2X");
	strcpy(card->shortname, "AT91 TLV320AIC2X");
	sprintf(card->longname, "Artila M501 with TI TLV320AIC2X");
        
	if ((err = snd_card_set_generic_dev(card)) < 0)
		goto nodev;

	if ((err = snd_card_register(card)) == 0) {
		printk( KERN_INFO "AT91 audio support initialized\n" );
		return 0;
	}
        
 nodev:
	snd_card_free(card);
	return err;
}

static void __exit at91_tlv320_exit(void)
{
	
	snd_card_free(at91_tlv320->card);
	
}

module_init(at91_tlv320_init);
module_exit(at91_tlv320_exit);

/* }}} */

/*
 * Local variables:
 * indent-tabs-mode: t
 * End:
 */
